#include "palclass.cpp"
#include "geoclass.cpp"
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <alloc.h>

template<class vtype1, class vtype2>
float rel_ang(vectortype<vtype1> vector1, vectortype<vtype2> vector2);

/************* classes ***************/
class platetype{
	public:
		pointtype<float> **corner;
		unsigned char colour;
		unsigned int numcorners;
		int avgz;

		platetype(){
			numcorners = 0;
			corner = NULL;
		}

		int calc_avgz(void);

		char init(int newnumcorners){
			freeme();
			numcorners = newnumcorners;
			corner = new pointtype<float> *[numcorners];
			if(corner) return 1;
			return 0;
		}

		void freeme(void){
			if(corner) delete corner;
			corner = NULL;
		}
};

int platetype::calc_avgz(void){
	int n;
	float z = 0;

	for(n = 0; n < numcorners; n++){
		z += corner[n]->z;
	}
	avgz = z / numcorners;

	return avgz;
}


class shapetype{
	public:
		vectortype<int> velocity;
		pointtype<int> position;
		int numplates, numcorners;
		platetype *plate;
		pointtype<float> *corner;

		shapetype(){ numplates = numcorners = 0; }
		void stretch(float xscale, float yscale, float zscale);
		void rotate(float xyang, float xzang, float yzang);
		void rotaround(pointtype<float> center, float xyang, float xzang, float yzang);
		void rotaround(pointtype<float> a, pointtype<float> b, float ang);
		void setcolour(unsigned char newcolour);
		void draw(void);
		void scale(float ratio);
		void freeme(void);
		char save(unsigned char *filename);
		char load(unsigned char *filename);
};

void shapetype::setcolour(unsigned char newcolour){
	int n;
	for(n = 0; n < numplates; n++)
		plate[n].colour = newcolour;
}

void shapetype::stretch(float xscale, float yscale, float zscale){
	int n;
	for(n = 0; n < numcorners; n++){
		corner[n].x *= xscale;
		corner[n].y *= yscale;
		corner[n].z *= zscale;
	}
}

void shapetype::freeme(void){
	int n;
	if(plate){
		for(n = 0; n < numplates; n++)
			plate[n].freeme();
		delete plate;
		plate = NULL;
	}
	if(corner){
		delete corner;
		corner = NULL;
	}
	numplates = numcorners = 0;
}

void shapetype::rotaround(pointtype<float> a, pointtype<float> b, float ang){
	//extern unsigned int maxz;
	vectortype<float> v1, v2;
	pointtype<float> c;
	float ang1, ang2;
	int n;

	a = a - position;
	b = b - position;
	v1.init(a.x, 0, a.z, b.x, 0, b.z);
	v2.init(0, 0, 0, 0, 0, 1000);
	ang1 = rel_ang(v1, v2);
	v1.init(0, a.y, a.z, 0, b.y, b.z);
	ang2 = rel_ang(v1, v2);

	for(n = 0; n < numcorners; n++){
		c.init(corner[n].x - a.x, corner[n].y - a.y, corner[n].z);

		c.rotate(0, ang1, ang2);
		c.rotate(ang, 0, 0);
		c.rotate(0, -ang1, -ang2);

		corner[n].init(c.x + a.x, c.y + a.y, c.z);
	}
}

void shapetype::rotaround(pointtype<float> center, float xyang, float xzang, float yzang){
	float SINE, COSINE;
	int n;
	float deltax, deltay, deltaz;

	SINE = sin(xyang);
	COSINE = cos(xyang);
	for(n = 0; n < numcorners; n++){
		deltax = corner[n].x - center.x;
		deltay = corner[n].y - center.y;
		corner[n].x = center.x + deltax * COSINE - deltay * SINE;
		corner[n].y = center.y + deltax * SINE + deltay * COSINE;
	}

	SINE = sin(xzang);
	COSINE = cos(xzang);
	for(n = 0; n < numcorners; n++){
		deltax = corner[n].x - center.x;
		deltaz = corner[n].z - center.z;
		corner[n].x = center.x + deltax * COSINE - deltaz * SINE;
		corner[n].z = center.z + deltax * SINE + deltaz * COSINE;
	}

	SINE = sin(yzang);
	COSINE = cos(yzang);
	for(n = 0; n < numcorners; n++){
		deltay = corner[n].y - center.y;
		deltaz = corner[n].z - center.z;
		corner[n].y = center.y + deltay * COSINE - deltaz * SINE;
		corner[n].z = center.z + deltay * SINE + deltaz * COSINE;
	}
}

void shapetype::rotate(float xyang, float xzang, float yzang){
	float SINE, COSINE;
	int n;
	float deltax, deltay, deltaz;

	SINE = sin(xyang);
	COSINE = cos(xyang);
	for(n = 0; n < numcorners; n++){
		deltax = corner[n].x;
		deltay = corner[n].y;
		corner[n].x = deltax * COSINE - deltay * SINE;
		corner[n].y = deltax * SINE + deltay * COSINE;
	}

	SINE = sin(xzang);
	COSINE = cos(xzang);
	for(n = 0; n < numcorners; n++){
		deltax = corner[n].x;
		deltaz = corner[n].z;
		corner[n].x = deltax * COSINE - deltaz * SINE;
		corner[n].z = deltax * SINE + deltaz * COSINE;
	}

	SINE = sin(yzang);
	COSINE = cos(yzang);
	for(n = 0; n < numcorners; n++){
		deltay = corner[n].y;
		deltaz = corner[n].z;
		corner[n].y = deltay * COSINE - deltaz * SINE;
		corner[n].z = deltay * SINE + deltaz * COSINE;
	}
}

void drawplate(pointtype<int> position, platetype plate);

void shapetype::draw(void){
	extern unsigned int maxz;
	float ratio;
	int n;
	pointtype<int> pos;
	platetype temp_plate;

	/******
		This is a quick bubble sort of the plates in order of
		which is closer (on average) to the user.
		This saves a lot of drawing time, and cuts down on the
		flicker if you're drawing directly to the screen.
	*******/
	for(n = 0; n < numplates; n++)
		plate[n].calc_avgz();

	for(n = 1; n < numplates; n++){
		if(plate[n].avgz < plate[n - 1].avgz){
			temp_plate = plate[n - 1];
			plate[n - 1] = plate[n];
			plate[n] = temp_plate;
			if(n > 1) n -= 2;
		}
	}
	/**********************/

	pos = position;
	ratio = (float)(maxz - pos.z + 50) / (float)maxz;
	pos.x = 160 + ((float)pos.x - 160) * ratio;
	pos.y = 100 + ((float)pos.y - 100) * ratio;
	for(n = 0; n < numplates; n++)
		drawplate(pos, plate[n]);
}


char shapetype::load(unsigned char *filename){
	char returnval = 1;
	FILE *fin;
	int n, m, l;

	fin = fopen(filename, "rb");
	if(fin){
		fread(&numcorners, 2, 1, fin);
		fread(&numplates, 2, 1, fin);
		corner = new pointtype<float>[numcorners];
		plate = new platetype[numplates];
		if(corner && plate){
			fread(corner, sizeof(pointtype<float>), numcorners, fin);

			for(n = 0; n < numplates; n++){
				fread(&plate[n].colour, 1, 1, fin);
				fread(&plate[n].numcorners, 2, 1, fin);
				if(plate[n].init(plate[n].numcorners)){
					for(m = 0; m < plate[n].numcorners; m++){
						fread(&l, 2, 1, fin);
						plate[n].corner[m] = &corner[l];
					}
				}else
					returnval = 0;

			}
		}else
			returnval = 0;
		fclose(fin);
		if(!returnval) freeme();

	}else returnval = 0;
	return returnval;
}

char shapetype::save(unsigned char *filename){
	FILE *fout;
	int n, m, l;

	fout = fopen(filename, "wb");
	if(fout){
		fwrite(&numcorners, 2, 1, fout);
		fwrite(&numplates, 2, 1, fout);
		fwrite(corner, sizeof(pointtype<float>), numcorners, fout);

		for(n = 0; n < numplates; n++){
			fwrite(&plate[n].colour, 1, 1, fout);
			fwrite(&plate[n].numcorners, 2, 1, fout);
			for(m = 0; m < plate[n].numcorners; m++){
				for(l = 0; &corner[l] != plate[n].corner[m]; l++);
				fwrite(&l, 2, 1, fout);
			}
		}
		fclose(fout);
		return 1;
	}else return 0;
}


/**************************** declarations *****************************/
void boing3d(void);
unsigned int readdepth(int x, int y);
void writedepth(int x, int y, unsigned int value);
void cleardepth(void);
void drawplate(pointtype<int> position, platetype plate);
void rotshape(shapetype *shape, int centerx, int centery, int centerz, float xyang, float xzang, float yzang);
void line3d(int x1, int y1, int z1, int x2, int y2, int z2, unsigned char colour);
void XZline(int x1, int z1, int x2, int z2, int drawy, unsigned char colour);
void polygon3d(pointtype<int> *point, int numpoints, unsigned char colour);
void update_screen(void);
void readmousediff(void);
char buttonpressed(void);
void initmouse(void);
char sgn(int val);
unsigned int counter(void);
void initscrn(char mode);
unsigned char csmode(void);
void cls(void);

/**************************** definitions ******************************/
#define tabsize 4
#define pi 3.1415926535897932384626433832795028841971693993751058209

/************************** Global Variables ***************************/
unsigned char far *mcgamem = (unsigned char far *)0xA0000000;
unsigned char far *depth1, far *depth2, far *vmem;
vectortype<int> mousediff;
vectortype<float> light_direction;
unsigned char **font = NULL, xfontsize, yfontsize;
palettetype palette;
unsigned int maxz = 1024;
int shading_range = 32, centerx = 160, centery = 100;


void boing3d(void){
	char memgood = 1, button_released = 1, bp;
	const int numshapes = 16;
	vectortype<int> gravity;
	shapetype shape[numshapes];
	int numtodraw = 6;
	unsigned int n, m, count;
	float ang = 0, angi = .05;
	for(n = 0; n < numshapes; n++)
		memgood &= shape[n].load("SPHERE2.S3D");

	do readmousediff(); while(mousediff.length());
	if(memgood){
		for(n = 0; n < numshapes; n++){
			switch(n % 2){
				case 0:
					for(m = 0; m < shape[n].numplates; m++){
						if(m % 2)
							shape[n].plate[m].colour += 32 * ((n + 1) % 6 + 1);
						else
							shape[n].plate[m].colour += 32 * ((n + 1) % 6);
					}
					break;
				case 1:
					for(m = 0; m < shape[n].numplates; m++){
						if(div(m, 12).quot % 2)
							shape[n].plate[m].colour += 32 * (n % 6 + 1);
						else
							shape[n].plate[m].colour += 32 * (n % 6);
					}
					break;

			}
			shape[n].stretch(1.5, 1.5, 1.5);
			shape[n].velocity.init(rand() % 4 + 2, rand() % 8 - 4, 5);
			if(rand() % 2) shape[n].velocity.reflectx();
			if(rand() % 2) shape[n].velocity.reflectz();
			shape[n].position.init(rand() % 320, rand() % 100 + 50, rand() % (maxz - 400) + 200);
			shape[n].rotate((rand() % 628) / 100, (rand() % 628) / 100, (rand() % 628) / 100);
		}

		gravity.init(0, 1, 0);
		initmouse();

		while(!kbhit()){
			while(count == counter());
			count = counter();
			cleardepth();
			cls();

			/******** handle mouse IO ********/
			bp = buttonpressed();
			if(!bp){
				button_released = 1;
			}else if(bp == 1){
				if(button_released){
					if(numtodraw < numshapes) numtodraw++;
					button_released = 0;
				}
			}else if(bp == 2){
				if(button_released){
					if(numtodraw > 0) numtodraw--;
					button_released = 0;
				}
			}
			readmousediff();

			if(mousediff.deltax || mousediff.deltay){
				light_direction.rotate(0, mousediff.deltax / 100., mousediff.deltay / 100.);
			}

			ang += angi;
			if(ang > 2 * pi) ang -= 2 * pi;

			for(n = 0; n < numtodraw; n++){
				/****** handle motion ******/
				shape[n].velocity += gravity;
				shape[n].position += shape[n].velocity;
				if(shape[n].position.x > 319 || shape[n].position.x < 0)
					shape[n].velocity.reflectx();
				if(shape[n].position.y > 199){
					shape[n].velocity.reflecty();
					shape[n].velocity -= gravity;
				}
				if(shape[n].position.z > maxz - 150 || shape[n].position.z < 50)
					shape[n].velocity.reflectz();

				/****** rotate and draw ********/
				shape[n].rotate(cos(ang) / 6., sin(ang) / 7., cos(ang) / 7.);
				shading_range = 4 + 32. * (float)(maxz - shape[n].position.z) / maxz;
				if(shading_range > 32.) shading_range = 32.;
				shape[n].draw();
			}

			update_screen();

		}
		while(kbhit()) getch();
	}

	for(m = 0; m < numshapes; m++){
		shape[m].freeme();
	}
}

void main(void){
	unsigned char smode;
	vmem = (unsigned char far *)farmalloc(64000);
	depth1 = (unsigned char far *)farmalloc(64000);
	depth2 = (unsigned char far *)farmalloc(64000);

	if(depth1 && depth2 && vmem){ /* memory successfully allocated */
		light_direction.init(0, 0, 0, -10, -10, -2);
		cls();
		smode = csmode();
		initscrn(0x13);
		if(csmode() == 0x13){ /* screen successfully initialized */
			palette.load("basics.plt");
			palette.apply();
			boing3d();

		}
		initscrn(smode);
	}else
		printf("Insufficient memory");
	if(depth1) farfree(depth1);
	if(depth2) farfree(depth2);
	if(vmem) farfree(vmem);
}

unsigned int readdepth(int x, int y){
	unsigned int mempos = x + (y << 8) + (y << 6);
	return (depth1[mempos] << 8) + depth2[mempos];
}

void writedepth(int x, int y, unsigned int value){
	unsigned int mempos = x + (y << 8) + (y << 6);
	depth1[mempos] = value >> 8;
	depth2[mempos] = value % 256;
}

void cleardepth(void){
	_fmemset(depth1, 255, 64000);
	_fmemset(depth2, 255, 64000);
}

void drawplate(pointtype<int> position, platetype plate){
	float ratio;
	unsigned char colour;
	vectortype<long> v1, v2, normal;
	pointtype<int> *drawpoint;
	int n, c1, c2, c3;

	drawpoint = new pointtype<int> [plate.numcorners];
	if(drawpoint){
		/**** alter colour to account for shading ****/
		/**
			This if tends to spread out the points used for calculating the
			normal.  Doing so improves the accuracy of it, thus improving
			the shading.
		**/
		if(plate.numcorners < 5){
			c1 = 0;
			c2 = 1;
			c3 = 2;
		}else{
			c1 = 0;
			c2 = plate.numcorners >> 2;
			c3 = plate.numcorners >> 1;
		}

		/** calculate the normal **/
		v1.init(plate.corner[c2]->x - plate.corner[c1]->x,
				plate.corner[c2]->y - plate.corner[c1]->y,
				plate.corner[c2]->z - plate.corner[c1]->z);

		v2.init(plate.corner[c3]->x - plate.corner[c1]->x,
				plate.corner[c3]->y - plate.corner[c1]->y,
				plate.corner[c3]->z - plate.corner[c1]->z);
		normal = v1 * v2;
		if(normal.deltaz < 0) normal *= -1;

		/** calculate the colour adjustment **/
		colour = plate.colour + (shading_range - 1) * (rel_ang(light_direction, normal) / pi);

		/**** add perspective ****/
		for(n = 0; n < plate.numcorners; n++){
			ratio = (float)(maxz - plate.corner[n]->z - position.z) / (float)maxz;
			drawpoint[n].init(plate.corner[n]->x * ratio + position.x,
							  plate.corner[n]->y * ratio + position.y,
							  plate.corner[n]->z + position.z);
		}

		/**** draw the polygon ****/
		polygon3d(drawpoint, plate.numcorners, colour);

		delete drawpoint;
	}
}

void rotshape(shapetype *shape, int centerx, int centery, int centerz, float xyang, float xzang, float yzang){
	float SINE, COSINE;
	int n;
	float deltax, deltay, deltaz;

	SINE = sin(xyang);
	COSINE = cos(xyang);
	for(n = 0; n < shape->numcorners; n++){
		deltax = shape->corner[n].x - (float)centerx;
		deltay = shape->corner[n].y - (float)centery;
		shape->corner[n].x = (float)centerx + deltax * COSINE - deltay * SINE;
		shape->corner[n].y = (float)centery + deltax * SINE + deltay * COSINE;
	}

	SINE = sin(xzang);
	COSINE = cos(xzang);
	for(n = 0; n < shape->numcorners; n++){
		deltax = shape->corner[n].x - (float)centerx;
		deltaz = shape->corner[n].z - (float)centerz;
		shape->corner[n].x = (float)centerx + deltax * COSINE - deltaz * SINE;
		shape->corner[n].z = (float)centerz + deltax * SINE + deltaz * COSINE;
	}

	SINE = sin(yzang);
	COSINE = cos(yzang);
	for(n = 0; n < shape->numcorners; n++){
		deltay = shape->corner[n].y - (float)centery;
		deltaz = shape->corner[n].z - (float)centerz;
		shape->corner[n].y = (float)centery + deltay * COSINE - deltaz * SINE;
		shape->corner[n].z = (float)centerz + deltay * SINE + deltaz * COSINE;
	}

}

void line3d(int x1, int y1, int z1, int x2, int y2, int z2, unsigned char colour){
	int n, deltax, deltay, deltaz;
	int sgndeltax, sgndeltay, sgndeltaz;
	int deltaxabs, deltayabs, deltazabs;
	int x, y, z;
	int drawx, drawy, drawz;

	deltax = x2 - x1;
	deltay = y2 - y1;
	deltaz = z2 - z1;
	deltaxabs = abs(deltax);
	deltayabs = abs(deltay);
	deltazabs = abs(deltaz);
	sgndeltax = sgn(deltax);
	sgndeltay = sgn(deltay);
	sgndeltaz = sgn(deltaz);
	x = deltayabs >> 1;
	y = deltaxabs >> 1;

	drawx = x1;
	drawy = y1;
	drawz = z1;

	if(drawx < 320 && drawx >= 0 && drawy < 200 && drawy >= 0){
		if(readdepth(drawx, drawy) > drawz){
			vmem[drawx + (drawy << 6) + (drawy << 8)] = colour;
			writedepth(drawx, drawy, drawz);
		}
	}

	if(deltaxabs >= deltayabs){
		z = deltaxabs >> 1;
		for(n = 0; n < deltaxabs; n++){
			y += deltayabs;
			if(y >= deltaxabs){
				y -= deltaxabs;
				drawy += sgndeltay;
			}

			z += deltazabs;
			while(z >= deltaxabs){
				z -= deltaxabs;
				drawz += sgndeltaz;
			}

			drawx += sgndeltax;
			if(drawx < 320 && drawx >= 0 && drawy < 200 && drawy >= 0){
				if(readdepth(drawx, drawy) > drawz){
					vmem[drawx + (drawy << 6) + (drawy << 8)] = colour;
					writedepth(drawx, drawy, drawz);
				}
			}
		}
	}else{
		z = deltayabs >> 1;
		for(n = 0; n < deltayabs; n++){
			x += deltaxabs;
			if(x >= deltayabs){
				x -= deltayabs;
				drawx += sgndeltax;
			}

			z += deltazabs;
			while(z >= deltayabs){
				z -= deltayabs;
				drawz += sgndeltaz;
			}

			drawy += sgndeltay;
			if(drawx < 320 && drawx >= 0 && drawy < 200 && drawy >= 0){
				if(readdepth(drawx, drawy) > drawz){
					vmem[drawx + (drawy << 6) + (drawy << 8)] = colour;
					writedepth(drawx, drawy, drawz);
				}
			}
		}
	}
}

void XZline(int x1, int z1, int x2, int z2, int drawy, unsigned char colour){
	int n, deltax, deltaz, sgndeltax, sgndeltaz;
	int deltaxabs, deltazabs, z, drawx, drawz;

	if(drawy >= 0 && drawy < 200){
		deltax = x2 - x1;
		deltaz = z2 - z1;
		deltaxabs = abs(deltax);
		deltazabs = abs(deltaz);
		sgndeltax = sgn(deltax);
		sgndeltaz = sgn(deltaz);
		z = deltaxabs >> 1;
		drawx = x1;
		drawz = z1;

		if(drawx >= 0 && drawx < 320){
			if(readdepth(drawx, drawy) > drawz){
				vmem[drawx + (drawy << 6) + (drawy << 8)] = colour;
				writedepth(drawx, drawy, drawz);
			}
		}

		for(n = 0; n < deltaxabs; n++){
			z += deltazabs;
			if(z >= deltaxabs){
				do{
					drawz += sgndeltaz;
					z -= deltaxabs;
				}while(deltaxabs < z);
			}

			drawx += sgndeltax;
			if(drawx >= 0 && drawx < 320){
				if(readdepth(drawx, drawy) > drawz){
					vmem[drawx + (drawy << 6) + (drawy << 8)] = colour;
					writedepth(drawx, drawy, drawz);
				}
			}
		}
	}
}

void polygon3d(pointtype<int> *point, int numpoints, unsigned char colour){
	char done = 0;
	int minpoint, n;
	int deltaxA, deltayA, deltazA, pointnumA;
	int deltaxB, deltayB, deltazB, pointnumB;
	int sgndeltaxA, sgndeltayA, sgndeltazA;
	int sgndeltaxB, sgndeltayB, sgndeltazB;
	int deltaxabsA, deltayabsA, deltazabsA;
	int deltaxabsB, deltayabsB, deltazabsB;
	int xA, zA, x1A, y1A, z1A, x2A, y2A, z2A;
	int xB, zB, x1B, y1B, z1B, x2B, y2B, z2B;
	int drawxA, drawyA, drawzA;
	int drawxB, drawyB, drawzB;

	if(numpoints > 2){
		/***** find the starting point ******/
		minpoint = 0;
		for(n = 1; n < numpoints; n++){
			if(point[n].y < point[minpoint].y)
				minpoint = n;
		}

		/******** initialize the starting points *********/
		x1A = point[minpoint].x;
		y1A = point[minpoint].y;
		z1A = point[minpoint].z;
		pointnumA = minpoint + 1;
		if(pointnumA < 0) pointnumA += numpoints;
		else pointnumA %= numpoints;
		x2A = point[pointnumA].x;
		y2A = point[pointnumA].y;
		z2A = point[pointnumA].z;
		deltaxA = x2A - x1A;
		deltayA = y2A - y1A;
		deltazA = z2A - z1A;
		deltaxabsA = abs(deltaxA);
		deltayabsA = abs(deltayA);
		deltazabsA = abs(deltazA);
		sgndeltaxA = sgn(deltaxA);
		sgndeltayA = sgn(deltayA);
		sgndeltazA = sgn(deltazA);
		xA = deltayabsA >> 1;
		zA = deltayabsA >> 1;
		drawxA = x1A;
		drawyA = y1A;
		drawzA = z1A;

		x1B = point[minpoint].x;
		y1B = point[minpoint].y;
		z1B = point[minpoint].z;
		pointnumB = minpoint - 1;
		if(pointnumB < 0) pointnumB += numpoints;
		else pointnumB %= numpoints;
		x2B = point[pointnumB].x;
		y2B = point[pointnumB].y;
		z2B = point[pointnumB].z;
		deltaxB = x2B - x1B;
		deltayB = y2B - y1B;
		deltazB = z2B - z1B;
		deltaxabsB = abs(deltaxB);
		deltayabsB = abs(deltayB);
		deltazabsB = abs(deltazB);
		sgndeltaxB = sgn(deltaxB);
		sgndeltayB = sgn(deltayB);
		sgndeltazB = sgn(deltazB);
		xB = deltayabsB >> 1;
		zB = deltayabsB >> 1;
		drawxB = x1B;
		drawyB = y1B;
		drawzB = z1B;

		/******** draw the polygon *********/
		while(!done){

			//line3d(drawxA, drawyA, drawzA, drawxB, drawyB, drawzB, colour);
			XZline(drawxA, drawzA, drawxB, drawzB, drawyA, colour);
			if(drawyA == y2A){
				if(y2A == y2B && x2A == x2B)
					done = 1;
				else{
					x1A = x2A;
					y1A = y2A;
					z1A = z2A;

					pointnumA++;
					if(pointnumA < 0) pointnumA += numpoints;
					else pointnumA %= numpoints;

					x2A = point[pointnumA].x;
					y2A = point[pointnumA].y;
					z2A = point[pointnumA].z;

					deltaxA = x2A - x1A;
					deltayA = y2A - y1A;
					deltazA = z2A - z1A;
					deltaxabsA = abs(deltaxA);
					deltayabsA = abs(deltayA);
					deltazabsA = abs(deltazA);
					sgndeltaxA = sgn(deltaxA);
					sgndeltayA = sgn(deltayA);
					sgndeltazA = sgn(deltazA);
					xA = deltayabsA >> 1;
					zA = deltazabsA >> 1;
					drawxA = x1A;
					drawyA = y1A;
					drawzA = z1A;
				}
			}

			xA += deltaxabsA;
			while(xA >= deltayabsA){
				xA -= (deltayabsA + !deltayabsA);
				drawxA += sgndeltaxA;
			}
			zA += deltazabsA;
			while(zA >= deltayabsA){
				zA -= (deltayabsA + !deltayabsA);
				drawzA += sgndeltazA;
			}
			drawyA += sgndeltayA;


			if(drawyB == y2B){
				if(y2A == y2B && x2A == x2B)
					done = 1;
				else{
					x1B = x2B;
					y1B = y2B;
					z1B = z2B;

					pointnumB--;
					if(pointnumB < 0) pointnumB += numpoints;
					else pointnumB %= numpoints;

					x2B = point[pointnumB].x;
					y2B = point[pointnumB].y;
					z2B = point[pointnumB].z;

					deltaxB = x2B - x1B;
					deltayB = y2B - y1B;
					deltazB = z2B - z1B;
					deltaxabsB = abs(deltaxB);
					deltayabsB = abs(deltayB);
					deltazabsB = abs(deltazB);
					sgndeltaxB = sgn(deltaxB);
					sgndeltayB = sgn(deltayB);
					sgndeltazB = sgn(deltazB);
					xB = deltayabsB >> 1;
					zB = deltazabsB >> 1;
					drawxB = x1B;
					drawyB = y1B;
					drawzB = z1B;
				}
			}

			xB += deltaxabsB;
			while(xB >= deltayabsB){
				xB -= (deltayabsB + !deltayabsB);
				drawxB += sgndeltaxB;
			}
			zB += deltazabsB;
			while(zB >= deltayabsB){
				zB -= (deltayabsB + !deltayabsB);
				drawzB += sgndeltazB;
			}
			drawyB += sgndeltayB;
		}
	}
}


void update_screen(void){
	_fmemcpy(mcgamem, vmem, 64000);
}

/**************************** Mouse I/O *****************************/
void readmousediff(void){
	int x, y;
	asm{
		mov ax, 0xB
		int 0x33
		mov x, cx
		mov y, dx
	}
	mousediff.x2 = mousediff.deltax = x;
	mousediff.y2 = mousediff.deltay = y;
	mousediff.x1 = mousediff.y1 = 0;
}

char buttonpressed(void){
	int returnval;
	asm{
		mov ax, 0x3
		int 0x33
		mov returnval, bx
	}
	return returnval;
}

void initmouse(void){
	asm{
		mov ax, 0
		int 0x33
	}
}

/****************************** Misc ******************************/
char sgn(int val){
	if(val > 0) return 1;
	return -(val < 0);
}

unsigned int counter(void){
	return peek(0x40, 0x6C);
}

/**** returns the relative angle between two vectors ****/
template<class vtype1, class vtype2>
float rel_ang(vectortype<vtype1> vector1, vectortype<vtype2> vector2){
	float a, b, c, returnval;
	double dxs, dys, dzs, dx, dy, dz;

	dx = (double)(vector1.deltax - vector2.deltax);
	dy = (double)(vector1.deltay - vector2.deltay);
	dz = (double)(vector1.deltaz - vector2.deltaz);
	dxs = fabs(dx * dx);
	dys = fabs(dy * dy);
	dzs = fabs(dz * dz);
	a = dxs + dys + dzs;//<--- NOTE: This is actually a^2
	b = vector1.length();
	c = vector2.length();
	if(b && c){
		returnval = (a - (b * b) - (c * c)) / -(2. * b * c);
		if(returnval > 1) returnval = 1;
		else if(returnval < -1) returnval = -1;
		return acos(returnval);
	}else
		return 0;

}


/****************************** Graphics ******************************/
void initscrn(char mode){
	asm{
		mov ah, 0
		mov al, mode
		int 0x10
	}
}

unsigned char csmode(void){
	unsigned char returnval;
	asm{
		mov ah, 0xF
		int 0x10
		mov returnval, al
	}
	return returnval;
}

void cls(void){
	_fmemset(vmem, 0, 64000);
}